#!/usr/bin/env python3

# This is free and unencumbered software released into the public domain.
#
# Anyone is free to copy, modify, publish, use, compile, sell, or
# distribute this software, either in source code form or as a compiled
# binary, for any purpose, commercial or non-commercial, and by any
# means.
#
# In jurisdictions that recognize copyright laws, the author or authors
# of this software dedicate any and all copyright interest in the
# software to the public domain. We make this dedication for the benefit
# of the public at large and to the detriment of our heirs and
# successors. We intend this dedication to be an overt act of
# relinquishment in perpetuity of all present and future rights to this
# software under copyright law.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
#
# For more information, please refer to <http://unlicense.org>


"""
This module provides some useful functions to get valid input from the
command line.
"""


# Functions
# ------------------------------------------------

def get_value(prompt, conv_func=None, check_func=None):
    """
    Tries to get a value from the user.

    If only "prompt" is given, this function is equal to the
    built-in function "input".

    If "conv_func" is given, the function will be called with the
    input-value. If the conversion fails, the input prompt will be
    shown again.

    If "check_func" is given, the function will be called with the
    converted input-value, if there's a "conv_func" else with the
    converted value. If the function returns "True", the converted-value will
    be returned.
    """
    if not prompt.endswith(" "):
        prompt += " "

    while True:
        temp = input(prompt)

        if conv_func is not None:
            try:
                temp = conv_func(temp)
            except:
                continue

        if check_func is None:
            break
        elif check_func(temp):
            break
    return temp


def get_int(prompt, check_func=None):
    """
    """
    return get_value(prompt, int, check_func)


def get_float(prompt, check_func=None):
    """
    """
    return get_value(prompt, float, check_func)


def choose(prompt, choices):
    """
    The user has to choose an element of *choices*.
    """
    # We add a list of the coices to the prompt. The items
    # index starts by 1.
    for i, choice in enumerate(choices):
        prompt += "\n{}. {}".format(i + 1, choice)
    prompt += "\n"

    # Let the user choose.
    while True:
        try:
            index = int(input(prompt)) - 1
            choice = choices[index]
        except KeyboardInterrupt:
            # This makes sure, that the user can still leave
            # the application by [ctrl + c].
            raise
        except:
            # Show the prompt again ...
            pass
        else:
            # Everything is ok: The user entered an integer
            # and the *index* is in the correct range: [0, len(choices)].
            # So we've got our value.
            break
    return (index, choice)


def ask(question, default=None):
    """
    Asks the user the question until he answers it with *yes* or *no*.

    Returns true if the user answered with yes.
    Returns *default* if the user only hit enter.
    """
    if default is None:
        prompt = question + " (yes|no) "
    elif default:
        prompt = question + " (YES|no) "
    else:
        prompt = question + " (yes|NO) "

    while True:
        temp = input(prompt)
        temp = temp.strip()
        temp = temp.upper()
        if default is not None and not temp:
            return default
        elif temp in ("Y", "YES"):
            return True
        elif temp in ("N", "NO"):
            return False



#!/usr/bin/env python3

# The MIT License (MIT)
#
# Copyright (c) 2014-2015 Benedikt Schmitt <benedikt@benediktschmitt.de>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
This module contains the :class:`ArgumentParser` class which wraps a
Python :class:`argparse.ArgumentParser` for the EMSM.
"""


# Modules
# ------------------------------------------------

# std
import argparse
import subprocess
import logging

# local
from .license_ import LICENSE
from .version import VERSION


# Data
# ------------------------------------------------

__all__ = [
    "LongHelpAction",
    "ArgumentParser"
    ]

log = logging.getLogger(__file__)


# Classes
# ------------------------------------------------

class LongHelpAction(argparse.Action):
    """
    Prints the *description* using less (if available) and exists.
    """

    def __init__(self, option_strings, description=None, dest=argparse.SUPPRESS,
                 default=argparse.SUPPRESS):
        """
        """
        self.description = description if description is not None else str()
        self.description.strip()

        super().__init__(
            option_strings = option_strings,
            dest = dest,
            default = default,
            nargs = 0,
            help = "shows the manual and exists"
            )
        return None

    def __call__(self, parser, namespace, values, option_string=None):
        """
        Prints the :attr:`description`.
        """
        # We print the docstring under linux using *less*, if available.
        was_printed = False
        try:
            less_proc = subprocess.Popen("less", stdin=subprocess.PIPE)
            try:
                less_proc.stdin.write(self.description.encode())
                was_printed = True
            except IOError as err:
                print(err)
            finally:
                less_proc.stdin.close()
                less_proc.wait()
        except (OSError, ValueError) as err:
            pass

        # Use print() if nothing else worked.
        if not was_printed:
            print(self.description)
            was_printed = True

        parser.exit()
        return None


class ArgumentParser(object):
    """
    Wraps an :class:`argparse.ArgumentParser` instance.

    This class handles the *root* EMSM argument parser. The root parser
    only has a few global EMSM commands like ``-w``, ``-s``. Each plugin has
    its own subparser:

    .. code-block:: bash

        $ foo@bar: minecraft [emsm args] (plugin_name) [plugin args]

    **Example:**

    .. code-block:: bash

        # Call the *worlds* plugin with the world *foo* as target.
        $ foo@bar: minecraft -w foo worlds --status
    """

    def __init__(self, app):
        """
        """
        self._app = app

        self._argparser = argparse.ArgumentParser(
            description = "Extendable Minecraft Server Manager (EMSM)",
            epilog = "Visit https://github.com/benediktschmitt/emsm for "\
                     "further information.",
            add_help=True
            )

        # This subparser contains the parsers for the plugins.
        self._plugin_subparsers = self._argparser.add_subparsers(
            title = "plugin",
            dest = "plugin",
            description = "The name of the plugin, you want to invoke."
            )

        # Contains and caches the parsed arguments.
        self._args = None
        return None

    def argparser(self):
        """
        Returns the wrapped :class:`argparse.ArgumentParser` instance.
        """
        return self._argparser

    def args(self, cache=True):
        """
        Parses (if not yet done) the command line arguments and returns a
        namespace object that contains the result.

        :param bool cache:
            If ``True``, and the arguments have already been parsed, the
            result of the previous parse is returned.

        .. seealso::

            * :meth:`argparse.ArgumentParser.parse_args`
        """
        if self._args is None or not cache:
            log.info("parsing arguments ...".format(self._args))

            self._args = self._argparser.parse_args()

            log.info("parsed arguments: {}".format(self._args))
        return self._args

    def plugin_parser(self, plugin_name):
        """
        Returns the subparser for the plugin with the name *plugin_name*.
        """
        return self._plugin_subparsers.add_parser(plugin_name)

    def setup(self):
        """
        Adds the global EMSM arguments to the root argument parser.

        This method has to be called, when the
        :class:`~emsm.core.worlds.WorldManager` and
        :class:`~emsm.core.server.ServerManager` have been loaded, since we
        require the names of the available worlds and server.
        """
        log.info("adding emsm arguments ...")

        # The selectable worlds.
        worlds_group = self._argparser.add_mutually_exclusive_group()
        worlds_group.add_argument(
            "-w", "--world",
            action = "append",
            dest = "worlds",
            metavar = "WORLD",
            choices = self._app.worlds().get_names(),
            default = list(),
            help = "Selects the world."
            )
        worlds_group.add_argument(
            "-W", "--all-worlds",
            action = "store_const",
            dest = "all_worlds",
            const = True,
            default = False,
            help = "Selects all available worlds."
            )

        # The selectable server.
        server_group = self._argparser.add_mutually_exclusive_group()
        server_group.add_argument(
            "-s", "--server",
            action = "append",
            dest = "server",
            metavar = "SERVER",
            choices = self._app.server().get_names(),
            default = list(),
            help = "Selects single server software."
            )
        server_group.add_argument(
            "-S", "--all-server",
            action = "store_const",
            dest = "all_server",
            const = True,
            default = False,
            help = "Selects all available server software."
            )
        return None



#!/usr/bin/env python3

# The MIT License (MIT)
#
# Copyright (c) 2014-2015 Benedikt Schmitt <benedikt@benediktschmitt.de>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


# Modules
# ------------------------------------------------

# std
import os
import logging
import configparser


# Backward compatibility
# ------------------------------------------------

try:
    FileNotFoundError
except NameError:
    FileNotFoundError = OSError


# Data
# ------------------------------------------------

__all__ = [
    "ConfigParser",
    "MainConfiguration",
    "ServerConfiguration",
    "WorldsConfiguration",
    "Configuration"
    ]

log = logging.getLogger(__file__)


# Classes
# ------------------------------------------------

class ConfigParser(configparser.ConfigParser):
    """
    Extends the standard Python :class:`configparser.ConfigParser` by some
    useful methods.

    :param str path:
        The path to the configuration file. This file is used, when you call
        :meth:`read` or :meth:`write`.
    """

    #: Written at the begin of the configuration file.
    _EPILOG = str()

    def __init__(self, path):
        """
        """
        super().__init__(
            allow_no_value = False,
            strict = True,
            empty_lines_in_values = False,
            interpolation=configparser.ExtendedInterpolation()
            )
        self._path = path
        return None

    def path(self):
        """
        Returns the path of the configuration file.
        """
        return self._path

    def read(self):
        """
        Reads the configuration from :meth:`path`.
        """
        try:
            with open(self._path, "r") as file:
                super().read_file(file)
        except (FileNotFoundError, IOError):
            pass
        return None

    def write(self):
        """
        Writes the configuration into :meth:`path`.
        """
        # Get the comment prefix.
        comment_prefix = self._comment_prefixes[0]
        comment_format = "{} {{}}".format(comment_prefix)

        # Convert the EPILOG to comment lines.
        epilog = map(lambda s: comment_format.format(s),
                     type(self)._EPILOG.split("\n"))
        epilog = "\n".join(epilog) + "\n\n"

        # Write the configuration into the file.
        with open(self._path, "w") as file:
            file.write(epilog)
            super().write(file)
        return None


class MainConfiguration(ConfigParser):
    """
    Handles the :file:`main.conf` configuration file.

    This file includes the configuration for the EMSM Application and the
    plugins.

    The EMSM owns the ``[emsm]`` section and each plugin has its own section
    with the plugin name.

    .. code-block:: ini

        [emsm]
        user = minecraft
        timeout = 0
        screenrc =

        [backups]
        include_server = ...
        # ...
    """

    _EPILOG = (
        "This file contains the settings for the EMSM core application and\n"
        "the plugins.\n"
        "\n"
        "The section of the EMSM looks like this per default:\n"
        "\n"
        "[emsm]\n"
        "user = minecraft\n"
        "timeout = -1\n"
        "screenrc = \n"
        "\n"
        "The configuration section of each plugin is titled with the plugins\n"
        "name."
        )

    def __init__(self, path):
        """
        """
        super().__init__(path)

        # Add the default configuration for the EMSM.
        self.add_section("emsm")
        self["emsm"]["user"] = "minecraft"
        self["emsm"]["timeout"] = "0"
        self["emsm"]["screenrc"] = ""
        return None


class ServerConfiguration(ConfigParser):
    """
    Handles the *server.conf* configuration file, which allows the user
    to overwrite the default EMSM settings for a server wrapper like
    the *url* or the *start command*.


    .. seealso::

        * :meth:`emsm.core.server.BaseServerWrapper.conf`
    """

    _EPILOG = (
        "[server name]\n"
        "url = string\n"
        "start_command = string\n"
        "\n"
        "The EMSM comes with tested default settings for each server.\n"\
        "so you should only overwrite these values, if you have to.\n"
        )


class WorldsConfiguration(ConfigParser):
    """
    Handles the *worlds.conf* configuration file, which defines the
    names and settings of all worlds managed by the EMSM.

    .. seealso::

        * :meth:`emsm.core.worlds.WorldWrapper.conf`
    """

    _EPILOG = (
        "[the world's name]\n"
        "stop_timeout = int\n"
        "stop_message = string\n"
        "stop_delay = int\n"
        "server = a server in server.conf\n"
        "\n"
        "Note, that some plugins may offer you some more options for\n"
        "a world, like *enable_initd*. Take a look at the plugins help page\n"
        "or documentation for further information.\n"
        )

    def __init__(self, path):
        """
        """
        super().__init__(path)

        # Populate the defaults section.
        defaults = self.defaults()
        defaults["stop_timeout"] = "10"
        defaults["stop_delay"] = "5"
        defaults["stop_message"] = "The server is going down.\n"\
                                   "Hope to see you soon."
        return None


class Configuration(object):
    """
    Manages all configuration files of an EMSM application
    object.

    .. seealso::

        * :meth:`emsm.core.application.Application.conf`
        * :meth:`emsm.core.paths.Pathsystem.conf_dir`
    """

    def __init__(self, app):
        """
        """
        self._app = app
        self._dir = app.paths().conf()

        self._main = MainConfiguration(os.path.join(self._dir, "main.conf"))
        self._server = ServerConfiguration(os.path.join(self._dir, "server.conf"))
        self._worlds = WorldsConfiguration(os.path.join(self._dir, "worlds.conf"))
        return None

    def main(self):
        """
        Returns the :class:`MainConfiguration`.
        """
        return self._main

    def server(self):
        """
        Returns the :class:`ServerConfiguration`.
        """
        return self._server

    def worlds(self):
        """
        Returns the :class:`WorldsConfiguration`.
        """
        return self._worlds

    def read(self):
        """
        Reads all configration files.
        """
        log.info("reading configuration ...")

        # Don't change the order!
        self._main.read()
        self._server.read()
        self._worlds.read()
        return None

    def write(self):
        """
        Saves all configuration values.
        """
        log.info("writing configuration ...")

        self._main.write()
        self._server.write()
        self._worlds.write()
        return None



#!/usr/bin/env python3

# The MIT License (MIT)
#
# Copyright (c) 2014-2015 Benedikt Schmitt <benedikt@benediktschmitt.de>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


# Modules
# ------------------------------------------------

# std
import os
import pwd
import grp
import sys
import logging
import traceback

# third party
import filelock
import colorama
import termcolor

# local
from . import argparse_
from . import base_plugin
from . import conf
from . import logging_
from . import paths
from . import plugins
from . import server
from . import worlds
from . import license_
from . import version


# Data
# ------------------------------------------------

__all__ = [
    "ApplicationException",
    "WrongUserError",
    "Application"
    ]

log = logging.getLogger(__file__)


# Exceptions
# ------------------------------------------------

class ApplicationException(Exception):
    """
    Base class for all exceptions in this module.
    """


class WrongUserError(ApplicationException):
    """
    Raised if the EMSM is executed by the wrong user.
    """

    def __init__(self, required_user):
        self.required_user = required_user
        return None

    def __str__(self):
        temp = "This script requires a user named '{}'."\
               .format(self.required_user)
        return temp


# Classes
# ------------------------------------------------

class Application(object):
    """
    This class manages the initialisation and the complete run process
    of the EMSM.

    An EMSM application should be executed in a code structure similar to this
    one:

    .. code-block:: python

        app = Application()
        try:
            app.setup()
            app.run()
        except Exception as err:
            app.handle_exception()
            raise
        finally:
            exit(app.finish())
    """

    def __init__(self, instance_dir):
        """
        """
        # The order of the initialisation is not trivial!
        self._paths = paths.Pathsystem(instance_dir)
        self._lock = filelock.FileLock(
            os.path.join(self._paths.instance(), "app.lock")
            )
        self._logger = logging_.Logger(self)

        self._conf = conf.Configuration(self)
        self._argparser = argparse_.ArgumentParser(self)

        self._worlds = worlds.WorldManager(self)
        self._server = server.ServerManager(self)
        self._plugins = plugins.PluginManager(self)

        # The exit code can be changed by plugins. This is useful
        # since a plugin should not throw a SystemExit exception.
        self._exit_code = 0
        return None

    def paths(self):
        """
        Returns the used :class:`~emsm.core.paths.Pathsystem` instance.
        """
        return self._paths

    def conf(self):
        """
        Returns the used :class:`~emsm.core.conf.Configuration` instance.
        """
        return self._conf

    def argparser(self):
        """
        Returns the EMSM :class:`~emsm.core.argparse_.ArgumentParser`, that is
        used internally.
        """
        return self._argparser

    def worlds(self):
        """
        Returns the used :class:`~emsm.core.worlds.WorldManager` instance.
        """
        return self._worlds

    def server(self):
        """
        Returns the used :class:`~emsm.core.server.ServerManager` instance.
        """
        return self._server

    def plugins(self):
        """
        Returns the used :class:`~emsm.core.plugins.PluginManager` instance.
        """
        return self._plugins

    def exit_code(self):
        """
        Returns the exit code of the application.

        .. seealso:: :meth:`set_exit_code`
        """
        return self._exit_code

    def set_exit_code(self, code):
        """
        Sets the exit code to *code*. This is the exit code, that is used for
        the Python :func:`exit` function.

        :raises TypeError:
            if *code* is not an int.
        :raises ValueError:
            if *code* < 0.

        .. seealso::

            * :meth:`exit_code`
            * :func:`exit`
        """
        if not isinstance(code, int):
            raise TypeError("*code* is not an int.")
        if code < 0:
            raise ValueError("*code* is < 0.")

        self._exit_code = code
        return None

    def _switch_user(self):
        """
        Switches the *uid* (user id) and *gui* (group id) of the current EMSM
        process to match the expected user defined in the :file:`main.conf`
        configuration file.

        :raises WrongUserError:
            if the rights could not be changed to the target user and group.

        .. seealso::

            * :meth:`emsm.core.conf.Configuration.main`
            * :func:`os.setuid`
            * :func:`os.setgid`
        """
        username = self._conf.main()["emsm"]["user"]

        user = pwd.getpwnam(username)
        group = grp.getgrgid(user.pw_gid)

        try:
            # Switch the group first.
            if os.getegid() != user.pw_gid:
                os.setgid(user.pw_gid)
                log.info("switched gid to '{}' ('{}')."\
                         .format(user.pw_gid, group.gr_name))

            # Switch the user.
            if os.geteuid() != user.pw_uid:
                os.setuid(user.pw_uid)
                log.info("switched uid to '{}' ('{}')."\
                         .format(user.pw_uid, user.pw_name))

        # We failed to switch the user and group.
        except OSError as err:
            log.critical(err, exc_info=True)
            raise WrongUserError(username)
        return None

    def handle_exception(self):
        """
        Checks :func:`sys.exc_info` if there is currently an uncaught exception
        and logs it.
        """
        exc_info = sys.exc_info()

        # Break if there is no exception that is currently handled.
        if None in exc_info:
            return None

        # Log all exceptions
        log.exception("uncaught exception:")

        # Handle the exception by creating a log entry and printing
        # a short error message.
        msg = "EMSM: Uncaught exception:\n"\
              " > Exception: {exc_type}\n"\
              " > Module     {exc_mod}\n"\
              " > Message:   {exc_msg}\n"\
              " > A full traceback can be found in the log file."\
              .format(exc_type = exc_info[0].__name__,
                      exc_mod = traceback.extract_tb(exc_info[2])[-1][0],
                      exc_msg = exc_info[1]
                      )
        msg = termcolor.colored(msg, "red")

        # Don't print the SystemExit exception.
        if not issubclass(exc_info[0], SystemExit):
            print(msg, file=sys.stderr)

        return None

    def setup(self):
        """
        Initialises all components of the EMSM.

        This method will block, until the EMSM filelock could be acquired or
        the configuration timeout value is reached.
        """
        log.info("----------")
        log.info("setting the EMSM {} up ...".format(version.VERSION))

        # Initialise colorama.
        colorama.init()

        # Read the configuration, so that we get to know some startup
        # parameters like the file lock *timeout* or the EMSM user.
        # Note, that the configuration wrappers define default values for
        # the EMSM. So the configuration files may not exist at this point and
        # we can call ``self._paths.create()`` later..
        self._conf.read()

        # Downgrade the privileges before doing anything else.
        self._switch_user()
        os.chdir(self._paths.instance())

        # Create the EMSM directories.
        # Note, that we must execute this after ``switch_user`` to make sure, the
        # files are owned by the EMSM user.
        self._paths.create()

        # Wait for the file lock to avoid running multiple EMSM applications
        # at the same time.
        log.info("waiting for the file lock ...")

        lock_timeout = self._conf.main()["emsm"].getint("timeout", 0)
        lock_timeout = lock_timeout if lock_timeout > 0 else None
        self._lock.acquire(lock_timeout, 0.05)

        # Now we have the file lock, so we can acquire the emsm.log file.
        self._logger.setup()

        # Reload the configuration again, since it may have changed while
        # waiting for the file lock.
        self._conf.read()

        self._plugins.setup()
        self._plugins.init_plugins()

        self._worlds.load_worlds()

        self._argparser.setup()
        return None

    def run(self):
        """
        Runs the plugins.

        .. seealso::

            * :meth:`emsm.core.plugins.PluginManager.run`
            * :meth:`emsm.core.plugins.PluginManager.finish`
        """
        # Parse the arguments.
        self._argparser.args(cache=False)

        # Dispatch the plugins.
        self._plugins.run()
        self._plugins.finish()

        # Save changes to the configuration that have been made during
        # execution.
        self._conf.write()
        return None

    def finish(self):
        """
        Performs some clean up and background stuff.

        :returns: :meth:`exit_code`

        .. note::

            Do not mix this method up with the
            :meth:`emsm.core.plugins.PluginManager.finish` method. These are not
            related.

        .. seealso::

            * :meth:`run`
            * :meth:`exit_code`
        """
        # Disable colorama.
        colorama.deinit()

        log.info("EMSM finished.")
        self._lock.release()
        return self._exit_code



#!/usr/bin/env python3

# The MIT License (MIT)
#
# Copyright (c) 2014-2015 Benedikt Schmitt <benedikt@benediktschmitt.de>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


# Modules
# ------------------------------------------------

# std
import os
import shlex
import shutil
import urllib.request
import logging
import subprocess
import re
import tempfile
import glob

# third party
import blinker
import yaml


# Backward compatibility
# --------------------------------------------------

if not hasattr(shlex, "quote"):
    # From the Python3.3 library
    _find_unsafe = re.compile(r'[^\w@%+=:,./-]', re.ASCII).search
    def _shlex_quote(s):
        """Return a shell-escaped version of the string *s*."""
        if not s:
            return "''"
        if _find_unsafe(s) is None:
            return s

        # use single quotes, and put single quotes into double quotes
        # the string $'b is then quoted as '$'"'"'b'
        return "'" + s.replace("'", "'\"'\"'") + "'"
    shlex.quote = _shlex_quote

try:
    FileNotFoundError
except:
    FileNotFoundError = OSError


# Data
# ------------------------------------------------

__all__ = ["ServerError",
           "ServerInstallationFailure",
           "ServerStatusError",
           "ServerIsOnlineError",
           "ServerIsOfflineError",
           "BaseServerWrapper",
           "ServerManager"
           ]

log = logging.getLogger(__file__)


# Exceptions
# --------------------------------------------------

class ServerError(Exception):
    """
    Base class for all exceptions in this module.
    """
    pass


class ServerInstallationFailure(ServerError):
    """
    Raised if a server installation failed.
    """

    def __init__(self, server, msg=None):
        self.server = server
        self.msg = str(msg)
        return None

    def __str__(self):
        temp = "The installation of the server '{}' failed."\
               .format(self.server.name())
        if self.msg is not None:
            temp += " " + self.msg
        return temp


class ServerStatusError(ServerError):
    """
    Raised if the server should be online/offline
    for an action but is offline/online.
    """

    def __init__(self, server, status, msg = str()):
        self.server = server
        self.status = status
        self.msg = str(msg)
        return None

    def __str__(self):
        if self.status:
            temp = "The server '{}' is online. {}"
        else:
            temp = "The server '{}' is offline. {}"
        temp = temp.format(self.server.name(), self.msg)
        return temp


class ServerIsOnlineError(ServerStatusError):
    """
    Raised if the server is online but should be offline.
    """

    def __init__(self, server, msg = str()):
        ServerStatusError.__init__(self, server, True, msg)
        return None


class ServerIsOfflineError(ServerStatusError):
    """
    Raised if the server is offline but should be online.
    """

    def __init__(self, server, msg = str()):
        ServerStatusError.__init__(self, server, False, msg)
        return None


# Classes
# ------------------------------------------------

class BaseServerWrapper(object):
    """
    Wraps a minecraft server (executable), NOT a world.

    The BaseServerWrapper is initialized using the options in the
    :file:`server.conf` configuration file.

    :param emsm.core.application.Application app:
        The parent EMSM application
    """

    @classmethod
    def name(self):
        """
        **ABSTRACT**

        The unique name of the server.

        **Example**:

        ``"vanilla 1.8"``
        """
        raise NotImplementedError()

    def __init__(self, app):
        """
        """
        log.info("initialising server '{}' ...".format(self.name()))

        self.__app = app

        # Absolute path to the directory which should contain all server
        # software.
        # For the vanilla server, this is only a single file
        # *minecraft_server.jar*. The forge server comes with some
        # subdirectories and libraries.
        self.__directory = app.paths().server_(self.name())

        # The configuration section in the *server.conf* configuration file.
        if not app.conf().server().has_section(self.name()):
            app.conf().server().add_section(self.name())
        self.__conf = app.conf().server()[self.name()]
        return None

    def directory(self):
        """
        Absolute path to the directory which contains all server software.
        """
        if not os.path.exists(self.__directory):
            os.makedirs(self.__directory)
        return self.__directory

    def exe_path(self):
        """
        **ABSTRACT**

        Absolute path to the server executable. This file is usually located
        in :meth:`directory`.
        """
        raise NotImplementedError()

    def conf(self):
        """
        Returns the configuration section in the *server.conf* configuration
        file.
        """
        return self.__conf

    def default_url(self):
        """
        **ABSTRACT**

        The URL where the server executable can be downloaded from.
        """
        raise NotImplementedError()

    def url(self):
        """
        Returns the url in :meth:`conf`, if available. Otherwise the value
        of :meth:`default_url`.
        """
        if "url" in self.conf():
            return self.conf().get("url")
        return self.default_url()

    def is_installed(self):
        """
        ``True`` if the executable has been downloaded and exists, otherwise
        ``False``.

        Per default, this method only checks if the :meth:`directory` is empty
        or not. It can be *overridden* for a more detailed check.
        """
        return bool(os.listdir(self.directory()))

    def is_online(self):
        """
        Returns ``True`` if at least one world is currently running with
        this server.
        """
        worlds = self.__app.worlds().get_by_pred(
            lambda w: w.server() is self and w.is_online()
            )
        return bool(worlds)

    def install(self):
        """
        **ABSTRACT**

        Installs the server by downloading it to :meth:`server`. If the
        server is already installed, nothing should happen.

        This method is called during the EMSM start phase if
        :meth:`is_installed` returns ``False``.

        :raises ServerInstallationFailure:
            * when the installation failed.
        """
        raise NotImplementedError()

    def reinstall(self):
        """
        Tries to reinstall the server. If the reinstallation fails, the
        old :meth:`server()` is restored and everything is like before.

        :raises ServerInstallationFailure:
            * when the installation failed.
        :raises ServerIsOnlineError:
            * when a world powered by this server software is online.
        """
        if self.is_online():
            raise ServerIsOnlineError(self)

        # Save the old directory in a temporary folder, so that we can restore
        # it if something fails.
        with tempfile.TemporaryDirectory() as tmp_dir:
            tmp_server_path = shutil.move(self.directory(), tmp_dir)

            # is_installed() returns now False.
            assert not self.is_installed()

            # So we can call install() again.
            try:
                self.install()
            except:
                # Clean up the installation target directory self.directory()
                # and move the old server path back.
                if os.path.exists(self.directory()):
                    if os.path.isdir(self.directory()):
                        shutil.rmtree(self.directory())
                    else:
                        # This is a relict of version 3, when not all server
                        # created a directory in ``server/``.
                        os.remove(self.directory())
                shutil.move(tmp_server_path, self.directory())

                # Reraise the original exception.
                raise
        return None

    def default_start_cmd(self):
        """
        **ABSTRACT**

        Returns the bash command string, that must be executed, to start the
        server.

        If there are paths in the returned command, they must be absolute.
        """
        raise NotImplementedError()

    def start_cmd(self):
        """
        Returns the value for *start_command* in :meth:`conf` if available
        and the :meth:`default_start_cmd` otherwise.
        """
        if "start_command" in self.conf():
            cmd = self.conf().get("start_command")
            cmd = cmd.format(
                server_exe = shlex.quote(self.exe_path()),
                server_dir = shlex.quote(self.directory())
            )
            return cmd
        else:
            return self.default_start_cmd()

    def translate_command(self, cmd):
        """
        **ABSTRACT**

        Translates the vanilla server command *cmd* to a command with the same
        meaning, but which can be understood by the server.

        **Example:**

        .. code-block:: python

            >>> # A BungeeCoord wrapper would do this:
            >>> bungeecord.translate_command("stop")
            "end"
            >>> bungeecord.translate_command("say Hello World!")
            "alert Hello World!"
        """
        return NotImplementedError()

    def log_path(cls, self):
        """
        **ABSTRACT**

        Returns the path of the server log file of a world.

        If a relative path is returned, the base path is the world
        directory.
        """
        raise NotImplementedError()

    def log_start_re(self):
        """
        **ABSTRACT**

        Returns a regex, that matches the first line in the log file,
        after a server restart.
        """
        raise NotImplementedError()

    def log_error_re(self):
        """
        **ABSTRACT**

        Returns a regex, that matches every line with a *severe* (critical)
        error.
        A severe error means, that the server does not run correct and needs
        to be restarted.
        """
        raise NotImplementedError()

    def world_address(self, world):
        """
        **ABSTRACT**

        Returns the address (ip, port) which is binded by the world.
        (None, None) should be returned, if the binding can not be retrieved.

        If the server is binded to all ip addresses, return the emtpy string
        ``""`` for the ip address.

        The port should be returned as integer. If it can not be retrieved,
        return None.
        """
        raise NotImplementedError()


# Vanilla
# '''''''

class VanillaBase(BaseServerWrapper):
    """
    Base class for all vanilla server versions.
    """

    def exe_path(self):
        return os.path.join(self.directory(), self.name())

    def default_start_cmd(self):
        return "java -jar {} nogui.".format(shlex.quote(self.exe_path()))

    def translate_command(self, cmd):
        return cmd

    def install(self):
        """
        """
        if self.is_installed():
            return None

        # Simply download the minecraft jar from mojang and copy the .jar in
        # the EMSM_ROOT/server directory.
        try:
            tmp_path, http_resp = urllib.request.urlretrieve(self.url())
        except Exception as err:
            raise ServerInstallationFailure(self, err)
        else:
            shutil.move(tmp_path, self.exe_path())
        return None

    def world_address(self, world):
        """
        """
        # Read the server.properties file in the world's directory.
        conf_path = os.path.join(world.directory(), "server.properties")
        try:
            with open(conf_path, "r") as file:
                conf = file.read()
        except (OSError, IOError) as err:
            port = None
            ip = None
        else:
            # Retrieve the values for *server-port* and *server-ip*.
            # Note, that we use the '^' and '$' chars, so that we only match
            # valid lines. If there is a syntax error in the configuration,
            # we will ignore the value and return None instead.
            port_re = "^server-port\s*=\s*([\d]{1,5})\s*$"
            port = re.findall(port_re, conf, re.MULTILINE)
            port = int(port[0]) if port else None

            ip_re = "^server-ip\s*=\s*(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})\s*$"
            ip = re.findall(ip_re, conf, re.MULTILINE)
            ip = ip[0] if ip else "localhost"
        return (ip, port)


class Vanilla_1_2(VanillaBase):

    @classmethod
    def name(self):
        return "vanilla 1.2"

    def default_url(self):
        return "http://s3.amazonaws.com/Minecraft.Download/versions/1.2.5/minecraft_server.1.2.5.jar"

    def log_path(self):
        return "./server.log"

    def log_start_re(self):
        return re.compile("^.*Starting minecraft server version 1\.2.*")

    def log_error_re(self):
        return re.compile(".* \[SEVERE\] .*", re.MULTILINE)


class Vanilla_1_3(VanillaBase):

    @classmethod
    def name(self):
        return "vanilla 1.3"

    def default_url(self):
        return "http://s3.amazonaws.com/Minecraft.Download/versions/1.3.2/minecraft_server.1.3.2.jar"

    def log_path(self):
        return "./server.log"

    def log_start_re(self):
        return re.compile("^.*Starting minecraft server version 1\.3.*")

    def log_error_re(self):
        return re.compile(".* \[SEVERE\] .*", re.MULTILINE)


class Vanilla_1_4(VanillaBase):

    @classmethod
    def name(self):
        return "vanilla 1.4"

    def default_url(self):
        return "http://s3.amazonaws.com/Minecraft.Download/versions/1.4.7/minecraft_server.1.4.7.jar"

    def log_path(self):
        return "./server.log"

    def log_start_re(self):
        return re.compile("^.*Starting minecraft server version 1\.4.*")

    def log_error_re(self):
        return re.compile(".* \[SEVERE\] .*", re.MULTILINE)


class Vanilla_1_5(VanillaBase):

    @classmethod
    def name(self):
        return "vanilla 1.5"

    def default_url(self):
        return "http://s3.amazonaws.com/Minecraft.Download/versions/1.5.2/minecraft_server.1.5.2.jar"

    def log_path(self):
        return "./server.log"

    def log_start_re(self):
        return re.compile("^.*Starting minecraft server version 1\.5.*")

    def log_error_re(self):
        return re.compile(".* \[SEVERE\] .*", re.MULTILINE)


class Vanilla_1_6(VanillaBase):

    @classmethod
    def name(self):
        return "vanilla 1.6"

    def default_url(self):
        return "https://s3.amazonaws.com/Minecraft.Download/versions/1.6.4/minecraft_server.1.6.4.jar"

    def log_path(self):
        return "./server.log"

    def log_start_re(self):
        return re.compile("^.*Starting minecraft server version 1\.6.*")

    def log_error_re(self):
        return re.compile(".* \[SEVERE\] .*", re.MULTILINE)


class Vanilla_1_7(VanillaBase):

    @classmethod
    def name(self):
        return "vanilla 1.7"

    def default_url(self):
        return "https://s3.amazonaws.com/Minecraft.Download/versions/1.7.10/minecraft_server.1.7.10.jar"

    def log_path(self):
        return "./logs/latest.log"

    def log_start_re(self):
        return re.compile("^.*Starting minecraft server version 1\.7.*")

    def log_error_re(self):
        return re.compile(".* \[SEVERE\] .*", re.MULTILINE)


class Vanilla_1_8(VanillaBase):

    @classmethod
    def name(self):
        return "vanilla 1.8"

    def default_url(self):
        return "https://s3.amazonaws.com/Minecraft.Download/versions/1.8.1/minecraft_server.1.8.1.jar"

    def log_path(self):
        return "./logs/latest.log"

    def log_start_re(self):
        return re.compile("^.*Starting minecraft server version 1\.8.*")

    def log_error_re(self):
        return re.compile(".* \[SEVERE\] .*", re.MULTILINE)

# MinecraftForger
# '''''''''''''''


class MinecraftForgeBase(BaseServerWrapper):
    """
    Base class for the minecraft forge server.

    Todo:
        * check if :meth:`log_error_re` is correct implemented.
    """

    def translate_command(self, cmd):
        return cmd

    def install(self):
        """
        """
        if self.is_installed():
            return None

        try:
            # We need to download the *installer* first.
            try:
                tmp_path, http_resp = urllib.request.urlretrieve(self.url())
            except Exception as err:
                raise ServerInstallationFailure(err)
            else:
                # Now, we have to run the installer.

                # Clear the server directory.
                shutil.rmtree(self.directory())
                if not os.path.exists(self.directory()):
                    os.makedirs(self.directory())

                os.chdir(self.directory())
                if not os.path.samefile(self.directory(), os.curdir):
                    msg = "Could not chdir to {}".format(self.directory())
                    raise ServerInstallationFailure(self, msg)

                sys_install_cmd = "java -jar {} --installServer"\
                                      .format(tmp_path)
                sys_install_cmd = shlex.split(sys_install_cmd)
                try:
                    p = subprocess.Popen(
                        sys_install_cmd,
                        stdout = subprocess.PIPE,
                        stderr = subprocess.PIPE
                        )

                    # Store the output of the installer in the logfiles.
                    out, err = p.communicate()
                    out = out.decode()
                    err = err.decode()
                    log.info(out)

                    # Check, if the installer exited with return code 0 and
                    # throw an exception if not.
                    if p.returncode:
                        msg = "Installer returned with '{}'."\
                              .format(p.returncode)
                        raise ServerInstallationFailure(self, msg)
                except Exception as err:
                    raise ServerInstallationFailure(self, err)
        except:
            # Try to undo the installation.
            if os.path.exists(self.directory()):
                shutil.rmtree(self.directory())

            raise
        return None

    def default_start_cmd(self):
        start_cmd = "java -jar {} nogui.".format(shlex.quote(self.exe_path()))
        return start_cmd


class MinecraftForge_1_6(MinecraftForgeBase, Vanilla_1_6):

    @classmethod
    def name(self):
        return "minecraft forge 1.6"

    def default_url(self):
        return "http://files.minecraftforge.net/minecraftforge/minecraftforge-installer-1.6.4-9.11.1.916.jar"

    def exe_path(self):
        filenames = [filename \
                     for filename in os.listdir(self.directory()) \
                     if re.match("^minecraftforge-universal-1\.6.*.jar$", filename)]
        filename = filenames[0]
        return os.path.join(self.directory(), filename)


class MinecraftForge_1_7(MinecraftForgeBase, Vanilla_1_7):

    @classmethod
    def name(self):
        return "minecraft forge 1.7"

    def default_url(self):
        return "http://files.minecraftforge.net/maven/net/minecraftforge/forge/1.7.10-10.13.2.1291/forge-1.7.10-10.13.2.1291-installer.jar"

    def exe_path(self):
        filenames = [filename \
                     for filename in os.listdir(self.directory()) \
                     if re.match("^forge-1\.7.*.jar$", filename)]
        filename = filenames[0]
        return os.path.join(self.directory(), filename)


class MinecraftForge_1_8(MinecraftForgeBase, Vanilla_1_8):

    @classmethod
    def name(self):
        return "minecraft forge 1.8"

    def default_url(self):
        return "http://files.minecraftforge.net/maven/net/minecraftforge/forge/1.8-11.14.0.1299/forge-1.8-11.14.0.1299-installer.jar"

    def exe_path(self):
        filenames = [filename \
                     for filename in os.listdir(self.directory()) \
                     if re.match("^forge-1\.8.*.jar$", filename)]
        filename = filenames[0]
        return os.path.join(self.directory(), filename)


# Bungeecord
# ''''''''''

class BungeeCordServerWrapper(BaseServerWrapper):
    """
    Wraps only the **latest** BungeeCord version.

    Unfortunetly, the BungeeCord server uses the git commit hash value
    as its version number. So it would be too much work to keep track of the
    versions.
    """

    @classmethod
    def name(self):
        return "bungeecord"

    def default_url(self):
        return "http://ci.md-5.net/job/BungeeCord/lastSuccessfulBuild/artifact/bootstrap/target/BungeeCord.jar"

    def exe_path(self):
        return os.path.join(self.directory(), self.name())

    def install(self):
        """
        """
        if self.is_installed():
            return None

        # Simply download the latest build and save it in the EMSM_ROOT/server
        # directory.
        try:
            tmp_path, http_resp = urllib.request.urlretrieve(self.url())
        except Exception as err:
            raise ServerInstallationFailure(err)
        else:
            shutil.move(tmp_path, self.exe_path())
        return None

    def default_start_cmd(self):
        return "java -jar {}".format(shlex.quote(self.exe_path()))

    def translate_command(self, cmd):
        cmd = cmd.strip()
        if cmd.startswith("say "):
            cmd = "alert " + cmd[len("say "):]
        elif cmd == "stop":
            cmd = "end"
        return cmd

    def log_path(self):
        return "./proxy.log.0"

    def log_start_re(self):
        return re.compile("^.*Enabled BungeeCord version git:.*")

    def log_error_re(self):
        return re.compile(".* \[SEVERE\] .*", re.MULTILINE)

    def world_address(self, world):
        """
        """
        # Try to read and parse the configuration.
        # Note that it may not exist, not readable or parseable.
        conf_path = os.path.join(world.directory(), "config.yml")
        try:
            with open(conf_path) as file:
                conf = yaml.load(file)
        except (OSError, IOError) as err:
            ip, port = (None, None)
        else:
            # Try to extract the ip and port.
            try:
                adr = conf["listeners"][0]["host"]
                ip, port = adr.split(":")
            except:
                ip, port = (None, None)
            else:
                # Check if the ip address is valid.
                ip = ip.strip()
                if not re.match("\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}", ip):
                    ip = None

                # Check if the server is bounded to all available ip addresses.
                if ip == "0.0.0.0":
                    ip = "localhost"

                # Check if the port is valid and convert it to an int.
                port = port.strip()
                port = int(port) if re.match("^\d{1,5}$", port) else None
        return (ip, port)


# Spigot (1.8+)
# '''''''''''''

class Spigot(BaseServerWrapper):
    """
    Wraps the **latest** Spigot version.
    """

    @classmethod
    def name(self):
        return "spigot latest"

    def default_url(self):
        return "https://hub.spigotmc.org/jenkins/job/BuildTools/lastSuccessfulBuild/artifact/target/BuildTools.jar"

    def build_dir(self):
        if "build_dir" in self.conf():
            tmp = self.conf().get("build_dir")
            tmp = os.path.expanduser(tmp)
            tmp = os.path.abspath(tmp)
            return tmp
        return tempfile.mkdtemp(prefix='spigotmc')

    def install(self):
        if self.is_installed():
            return None

        # Download and build in /tmp
        build_dir = self.build_dir()

        log.info("Installing spigot ...")
        log.info("- Building in '{}' ...".format(build_dir))
        try:
            # Download the build tools.
            try:
                buildtools, http_resp = urllib.request.urlretrieve(
                    self.url(), os.path.join(build_dir, "BuildTools.jar")
                    )
            except Exception as err:
                raise ServerInstallationFailure(self, err)

            log.info("- BuildTools: '{}' ...".format(buildtools))

            # Run the BuildTools.jar file.
            with subprocess.Popen(
                ["java", "-jar", buildtools],
                cwd = build_dir,
                stdout = subprocess.PIPE,
                stderr = subprocess.PIPE
            ) as proc:
                # Store the output of the installer in the logfiles.
                out, err = proc.communicate()
                out = out.decode()
                err = err.decode()
                log.info(out)
                log.error(err)

                # Check, if the installer exited with return code 0 and
                # throw an exception if not.
                if proc.returncode:
                    msg = "Installer returned with '{}'."\
                          .format(proc.returncode)
                    raise ServerInstallationFailure(self, msg)

            # Move the built files to *exe_path()*.
            jarfiles = glob.glob(
                os.path.join(build_dir, "Spigot/Spigot-Server/target/spigot-*.jar")
                )
            if len(jarfiles) > 0:
                jarfile = jarfiles[0]
                shutil.move(jarfile, self.exe_path())
            else:
                msg = "Could not find built spigot-*.jar file."
                raise ServerInstallationFailure(self, msg)

        finally:
            # Remove the build directory, if it's only a temporary dir.
            if "build_dir" not in self.conf():
                log.info("- Removing build directory {}".format(build_dir))
                shutil.rmtree(build_dir)
        return None

    def exe_path(self):
        return os.path.join(self.directory(), self.name())

    def default_start_cmd(self):
        return "java -jar {}".format(shlex.quote(self.exe_path()))

    def log_path(self):
        return "./logs/latest.log"

    def log_start_re(self):
        return re.compile("^.*Starting minecraft server version .*")

    def log_error_re(self):
        """
        Todo: Check if this regex is correct and matches an error line.
        """
        return re.compile(".*/SEVERE\].*", re.MULTILINE)

    def translate_command(self, cmd):
        return cmd

    def world_address(self, world):
        """
        """
        # Read the server.properties file in the world's directory.
        conf_path = os.path.join(world.directory(), "server.properties")
        try:
            with open(conf_path, "r") as file:
                conf = file.read()
        except (OSError, IOError) as err:
            port = None
            ip = None
        else:
            # Retrieve the values for *server-port* and *server-ip*.
            # Note, that we use the '^' and '$' chars, so that we only match
            # valid lines. If there is a syntax error in the configuration,
            # we will ignore the value and return None instead.
            port_re = "^server-port\s*=\s*([\d]{1,5})\s*$"
            port = re.findall(port_re, conf, re.MULTILINE)
            port = int(port[0]) if port else None

            ip_re = "^server-ip\s*=\s*(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})\s*$"
            ip = re.findall(ip_re, conf, re.MULTILINE)
            ip = ip[0] if ip else "localhost"
        return (ip, port)


# MC-Server
# '''''''''

class MCServer(BaseServerWrapper):
    pass


# Server DB
# ------------------------------------------------

class ServerManager(object):
    """
    Manages all server wrappers, owned by an EMSM application.

    The ServerManager helps to avoid double instances of the same server
    wrapper.
    """

    def __init__(self, app):
        """
        """
        self._app = app

        # Maps *server.name()* to *server*
        self._server = dict()
        self.__add_emsm_wrapper()
        return None

    def __add_emsm_wrapper(self):
        """
        Loads all default EMSM server wrappers. These are all server wrappers
        defined in this module.
        """
        # We only add the complete wrappers. These wrappers have all
        # virtual (abstract) methods implemented.
        wrappers = [
            Vanilla_1_2,
            Vanilla_1_3,
            Vanilla_1_4,
            Vanilla_1_5,
            Vanilla_1_6,
            Vanilla_1_7,
            Vanilla_1_8,
            MinecraftForge_1_6,
            MinecraftForge_1_7,
            MinecraftForge_1_8,
            BungeeCordServerWrapper,
            Spigot,
            ]

        for wrapper in wrappers:
            self.add(wrapper)
        return None

    def add(self, server_class):
        """
        Makes the *server_class* visible to this manager. The class
        must implement all abstract methods of :class:`BaseServerWrapper` or
        unexpected errors may occure.

        :raises TypeError:
            if *server_class* does not inherit :class:`BaseServerWrapper`
        :raises ValueError:
            if another wrapper with the :meth:`~BaseServerWrapper.name()` of
            *server_class* has already been registered.
        """
        if not issubclass(server_class, BaseServerWrapper):
            raise TypeError("server_class has to inherit from BaseServerWrapper")

        if server_class.name() in self._server:
            raise ValueError("another server with the name '{}' has already "\
                             "been registered.".format(server_class.name()))

        # Create a new instance of the server wrapper.
        self._server[server_class.name()] = server_class(self._app)
        return None

    def get(self, servername):
        """
        Returns the :class:`ServerWrapper` with the name *servername* and
        ``None``, if there is not such a server.
        """
        return self._server.get(servername)

    def get_all(self):
        """
        Returns a list with all loaded :class:`ServerWrapper`.
        """
        return list(self._server.values())

    def get_by_pred(self, pred=None):
        """
        Almost equal to:

        .. code-block:: python

            >>> filter(pred, ServerManager.get_all())
            ...
        """
        return list(filter(pred, self._server.values()))

    def get_selected(self):
        """
        Returns all server that have been selected per command line argument.
        """
        args = self._app.argparser().args()

        selected_server = args.server
        all_server = args.all_server

        if all_server:
            return list(self._server.values())
        else:
            return [self._server[server] for server in selected_server]

    def get_names(self):
        """
        Returns a list with the names of all server.
        """
        return list(self._server.keys())



#!/usr/bin/env python3

# The MIT License (MIT)
#
# Copyright (c) 2014-2015 Benedikt Schmitt <benedikt@benediktschmitt.de>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
About
-----

This plugin provides a user interface for the server wrapper. It handles
the server files and their configuration parameters easily.

Download
--------

You can find the latest version of this plugin in the **EMSM**
`GitHub repository <https://github.com/benediktschmitt/emsm>`_.

Configuration
-------------

.. code-block:: ini

    [worlds]
    default_log_start = 0
    default_log_limit = 10
    open_console_delay = 1
    send_command_timeout = 10

**default_log_start**

    Is the first line of the log, that is printed. Can be overwritten by a
    command line argument.

**default_log_limit**

    Is the default number of log lines, that is printed at once. This
    value can be overwritten by a command line argument too.

**open_console_delay**

    Time between printing the WARNING and opening the console.

**send_command_timeout**

    Maximum time waited for the response of the minecraft server,
    if the ``--verbose-send`` command is used.

Arguments
---------

.. option:: --address

    Prints the binding (ip, port) of the world.

.. option:: --configuration

    Prints the section of the world in the :file:`worlds.conf`.

.. option:: --directory

    Prints the directory path that contains the world.

.. option:: --log

    Prints the log.

.. option:: --log-start LINE

    The first line of the log that is printed. If *'-10'* (with quotes!), the
    10th last line will be the first line that is printed.

.. option:: --log-limit LINES

    Limits the number of printed lines.

.. option:: --pid

    Prints the PID of the screen session that runs the server.

.. option:: --status

    Prints the status of the world (online or offline).

.. option:: --send CMD

    Sends the command to the world.

    .. note:: Escaping commands with **spaces**

        If you want to send a command like ``say Hello players!``, you have to
        escape it.

        .. code-block:: bash

            minecraft -W worlds --send 'say Hello players!'

.. option:: --verbose-send CMD

    Sends the command to the server and prints the echo in the logfiles.

.. option:: --console

    Opens the server console.

.. option:: --start

    Starts the world

.. option:: --stop

    .. warning::

        Stopping the world not using the dedicated commands, will **not**
        call the **event dispatcher** and may cause bugs.

    Stops the world

.. option:: --force-stop

   Like --stop, but kill the processes if the world is still online
   after the smooth stop.

.. option:: --kill

    .. warning::

        Using this command can cause data loss.

    Kills the process of the world.

.. option:: --restart

    Restarts the world. If the world is offline, the world will be started.

.. option:: --force-restart

    Like --restart, but forces the stop of the world if necessairy.

.. option:: --uninstall

    Removes the world and its configuration.

Examples
---------

.. code-block:: bash

    # Start all worlds:
    $ minecraft -W worlds --start

    # Send a command to the server and print the console output:
    $ minecraft -W worlds --verbose-send list
    $ minecraft -W worlds --verbose-send '"say Use more TNT!"'

    # Print the log of the world *foo*:
    $ minecraft -w foo worlds --log
    $ minecraft -w foo worlds --log-start '-20'
    $ minecraft -w foo worlds --log-limit 5
    $ minecraft -w foo worlds --log-start '-50' --log-limit 10

    # Open the console of a running world
    $ minecraft -w bar worlds --console

    ...
"""


# Modules
# --------------------------------------------------

# std
import os
import sys
import time

# third party
import termcolor

# emsm
import emsm
from emsm.core.base_plugin import BasePlugin


# Data
# --------------------------------------------------

PLUGIN = "Worlds"


# Classes
# --------------------------------------------------
class MyWorld(object):
    """
    Wraps an EMSM WorldWrapper instance. :)
    """

    def __init__(self, app, world):
        """
        """
        self._app = app
        self._world = world
        return None

    def world(self):
        """
        Returns the wrapped WorldWrapper object.
        """
        return self._world

    def print_address(self):
        """
        Prints the remote address (ip, port) the server is binded to.

        See also:
            * WorldWrapper.address()
        """
        ip, port = self._world.address()

        print(termcolor.colored("{}:".format(self._world.name()), "cyan"))
        if ip and port:
            print("\t", "{}:{}".format(ip, port))
        elif (not ip) and port:
            print("\t", "*:{}".format(port))
        elif ip and (not port):
            print("\t", "{}:{}".format(ip, termcolor.colored("?????", "red")))
        else:
            print("\t", termcolor.colored("error:", "red"),
                  "unable to retrieve binding.")
        return None

    def print_conf(self):
        """
        Prints the configuration of the world.

        See also:
            * WorldWrapper.conf()
        """
        conf = self._world.conf().items()

        print(termcolor.colored("{}:".format(self._world.name()), "cyan"))
        for key, value in sorted(conf):
            # Make sure, the indent of multiline values is correct.
            value = value.replace("\n", "\n\t\t")

            print("\t", key, "=", value)
        return None

    def print_directory(self):
        """
        Prints the path to the world's directory.

        See also:
            * WorldWrapper.directory()
        """
        print(termcolor.colored("{}:".format(self._world.name()), "cyan"))
        print("\t", self._world.directory())
        return None

    def print_latest_log(self, start_line=0, line_limit=20):
        """
        Prints the latest log of the world.

        Parameters:
            * start_line
                The number of the line that is printed first. If < 0, we start
                counting at the end.
            * line_limit
                The number of log lines that is printed. *0* means no limit.

        Example:
            >>> # Print the first 20 lines since the server start.
            >>> world.print_latest_log(0, 20)
            ...
            >>> # Print the latest 20 lines..
            >>> world.print_latest_log(-1, 20)
            >>> ...

        See also:
            * WorldWrapper.latest_log()
        """
        log = self._world.latest_log()
        log = log.split("\n")

        # Make sure that the limits are valid.
        # 0 <= start <= size
        # limit <= size - start
        num_lines = len(log)

        if start_line >= 0:
            start_line = min(start_line, num_lines)
        else:
            # start_line < 0
            start_line = max(0, num_lines + start_line)

        if line_limit != 0:
            line_limit = min(line_limit, num_lines - start_line)
            end_line = start_line + line_limit
        else:
            end_line = num_lines

        # Print the log section.
        tmp = termcolor.colored(self._world.name(), "cyan") + " - " +\
              termcolor.colored("lines {}-{}/{}".format(start_line + 1, end_line, num_lines), "green") +\
              ":"
        print(tmp)

        for line in range(start_line, end_line):
            tmp = "#{line_no:>8} | {line}"
            tmp = tmp.format(
                line_no = termcolor.colored(str(start_line + line + 1)),
                line = log[line]
                )
            print("\t", tmp)
        return None

    def print_pids(self):
        """
        Prints the pids of the screen sessions that run the minecraft server
        for this world.

        See also:
            * WorldWrapper.pids()
        """
        pids = self._world.pids()

        print(termcolor.colored("{}:".format(self._world.name()), "cyan"))
        if not pids:
            print("\t", "- offline -")
        else:
            for pid in pids:
                print("\t", pid)
        return None

    def print_status(self, status=None):
        """
        Prints the current status (*offline* or *online*) of the world.

        See also:
            * WorldWrapper.is_online()
        """
        print(termcolor.colored("{}:".format(self._world.name()), "cyan"))
        if self._world.is_online():
            print("\t", termcolor.colored("online", "green"))
        else:
            print("\t", termcolor.colored("offline", "red"))
        return None

    def send_command(self, cmd):
        """
        Sends the command *cmd* to the server.

        Params:
            * cmd
                The minecraft server command to send.

        See also:
            * WorldWrapper.send_command()
        """
        print(termcolor.colored("{}:".format(self._world.name()), "cyan"))
        try:
            self._world.send_command(cmd)
        except emsm.core.worlds.WorldIsOfflineError:
            print("\t", termcolor.colored("error:", "red"), "the world is offline")
        else:
            print("\t", "done.".format(cmd))
        return None

    def verbose_send_command(self, cmd, timeout):
        """
        Sends the command to the world and prints the output.

        Params:
            * cmd
                The minecraft command that is sent to the server.
            * timeout
                Maximum time waited for the server response.

        See also:
            * WorldWrapper.send_command_get_output()
        """
        print(termcolor.colored("{}:".format(self._world.name()), "cyan"))
        try:
            output = self._world.send_command_get_output(
                server_cmd=cmd, timeout=timeout
                )
        except emsm.core.worlds.WorldIsOfflineError:
            print("\t", termcolor.colored("error:", "red"), "the world is offline")
        except emsm.core.worlds.WorldCommandTimeout:
            print("\t", termcolor.colored("error:", "red"), "the world did not react")
        else:
            for line in output.split("\n"):
                print("\t", line)
        return None

    def open_console(self, delay=1):
        """
        Opens the screen session of the world's server process.

        Parameters:
            * delay
                The time between the warning message and the actual
                *open* command.

        See also:
            * WorldWrapper.open_console()
        """
        print(termcolor.colored("{}:".format(self._world.name()), "cyan"))
        if self._world.is_online():
            print("\t", termcolor.colored("note:   ", "yellow"),
                  "Press [ctrl + a + d] (in this order) to detach from the console."
                  )
            print("\t", termcolor.colored("warning:", "red"),
                  "When you stop the server in the session, EMSM may behave"\
                  "unexpected."
                  )
            time.sleep(delay)

            # No need to catch the WorldIsOfflineError, since we check the
            # status.
            self._world.open_console()
        else:
            print("\t", termcolor.colored("error:", "red"), "the world is offline.")
        return None

    def start(self):
        """
        Starts the server.

        See also:
            * WorldWrapper.start()
        """
        print(termcolor.colored("{}:".format(self._world.name()), "cyan"))
        try:
            self._world.start()
        except emsm.core.worlds.WorldStartFailed:
            print("\t", termcolor.colored("error:", "red"), "the world could not be started.")
        else:
            print("\t", "the world is now", termcolor.colored("online", "green"))
        return None

    def kill_processes(self):
        """
        Kills all server processes of the world.

        See also:
            * WorldWrapper.kill_processes()
        """
        print(termcolor.colored("{}:".format(self._world.name()), "cyan"))
        try:
            self._world.kill_processes()
        except emsm.core.worlds.WorldStopFailed:
            print("\t", termcolor.colored("error:", "red"), "the world could not be stopped.")
        else:
            print("\t", "the world is now", termcolor.colored("offline", "red"))
        return None

    def stop(self, force_stop=False):
        """
        Stops the server.

        Parameters:
            * force_stop
                If *true*, the ``kill processess`` is called if the
                server did not react on the normal shutdown command.

        See also:
            * WorldWrapper.stop()
        """
        print(termcolor.colored("{}:".format(self._world.name()), "cyan"))
        try:
            self._world.stop(force_stop=force_stop)
        except emsm.core.worlds.WorldStopFailed:
            if force_stop:
                print("\t", termcolor.colored("error:", "red"), "the world could not be stopped.")
            else:
                print("\t", termcolor.colored("error:", "red"), "the world could not be stopped.")
                print("\t", "       try: *--force-stop*")
        else:
            print("\t", "the world is now", termcolor.colored("offline", "red"))
        return None

    def restart(self, force_restart=False):
        """
        Restarts the world.

        Parameters:
            * force_restart
                If true, the stop of the world is force by *kill processes*
                if necessairy.

        See also:
            * WorldWrapper.restart()
        """
        print(termcolor.colored("{}:".format(self._world.name()), "cyan"))
        try:
            self._world.restart(force_restart=force_restart)
        except emsm.core.worlds.WorldStopFailed:
            if force_restart:
                print("\t", termcolor.colored("error:", "red"), "the world could not be stopped.")
            else:
                print("\t", termcolor.colored("error:", "red"), "the world could not be stopped.")
                print("\t", "       try: *--force-restart*")
        except emsm.core.worlds.WorldStartFailed:
            print("\t", termcolor.colored("error:", "red"), "the world could not be restarted.")
        else:
            print("\t", "the world has been", termcolor.colored("restarted.", "yellow"))
        return None

    def uninstall(self):
        """
        Removes the world from EMSM.

        See also:
            * WorldWrapper.uninstall()
        """
        print(termcolor.colored("{}:".format(self._world.name()), "cyan"))

        # Make sure, that the user wants to remove the world.
        prompt = "\t Are you sure, that you want to remove the world?"

        if emsm.core.lib.userinput.ask(prompt):
            self._world.uninstall()
            print("\t", "The world has been removed.")
        else:
            print("\t", "- aborted -")
        return None


class Worlds(BasePlugin):

    VERSION = "4.0.0-beta"

    DESCRIPTION = __doc__

    def __init__(self, application, name):
        """
        """
        BasePlugin.__init__(self, application, name)

        self.setup_conf()
        self.setup_argparser()
        return None

    def setup_conf(self):
        """
        Loads the configuration values and makes sure they have a valid value.
        """
        conf = self.conf()

        # Load the values.
        self._default_log_start = conf.getint(
            "default_log_start", 0)
        self._default_log_limit = conf.getint(
            "default_log_limit", 10)
        self._default_open_console_delay = conf.getint(
            "open_console_delay", 1)
        self._default_send_command_timeout = conf.getint(
            "send_command_timeout", 10)

        # And store the *used* values in the configuration.
        # This makes sense since we initialise the configuration section
        # this way and overwrite invalid values catch by *get()* above.
        conf["default_log_start"] = str(self._default_log_start)
        conf["default_log_limit"] = str(self._default_log_limit)
        conf["open_console_delay"] = str(self._default_open_console_delay)
        conf["send_command_timeout"] = str(self._default_send_command_timeout)
        return None

    def setup_argparser(self):
        """
        Adds the accepted arguments to the argpaser of this plugin.
        """
        parser = self.argparser()
        parser.description = "Manage and interact with a world."

        # Todo: We support only one action per run. So we should put all
        #       arguments in a mutually exclusive group. The only problem
        #       with python 3.4 was, the *log* group. The log command can
        #       be combined with the log-start and log-limit command.

        # Information about the world.
        parser.add_argument(
            "--address",
            action = "count",
            dest = "worlds_address",
            help = "Prints the binding (ip, port) of the world."
            )
        parser.add_argument(
            "--configuration",
            action = "count",
            dest = "configuration",
            help = "Prints the configuration section."
            )
        parser.add_argument(
            "--directory",
            action = "count",
            dest = "directory",
            help = "Prints the path to the world's directory."
            )

        # Log group
        log_group = parser.add_argument_group(title="log")
        log_group.add_argument(
            "--log",
            action = "count",
            dest = "log",
            help = "Prints the log."
            )
        log_group.add_argument(
            "--log-start",
            action = "store",
            dest = "log_start",
            type = int,
            help = "First printed line of the log. "\
            "('\"-2\"' starts with the 2nd last line)"
            )
        log_group.add_argument(
            "--log-limit",
            action = "store",
            dest = "log_limit",
            type = int,
            help = "The number of lines that will be printed."
            )

        # XXX: I need a name for that group
        # of arguments.
        console_group = parser.add_argument_group(title="console")
        console_group.add_argument(
            "--send",
            action = "store",
            dest = "send",
            metavar = "COMMAND",
            help = "Sends the command to the world. E.g.: '\"say Hello\"'"
            )
        console_group.add_argument(
            "--verbose-send",
            action = "store",
            dest = "verbose_send",
            metavar = "COMMAND",
            help = "Sends the command to the world prints the log echo."
            )
        console_group.add_argument(
            "--console",
            action = "count",
            dest = "console",
            help = "Opens the server console of the world."
            )

        # Status changes
        status_group = parser.add_argument_group(title="status")
        status_group.add_argument(
            "--pid",
            action = "count",
            dest = "pid",
            help = "Prints the pid of the server that runs the world."
            )
        status_group.add_argument(
            "--status",
            action = "count",
            dest = "status",
            help = "Prints the status of the world."
            )
        status_group.add_argument(
            "--start",
            action = "count",
            dest = "start",
            help = "Starts the world."
            )
        status_group.add_argument(
            "--stop",
            action = "count",
            dest = "stop",
            help = "Stops the world."
            )
        status_group.add_argument(
            "--force-stop",
            action = "count",
            dest = "force_stop",
            help = "Like --stop, but kills the processes "\
            "if the smooth stop fails."
            )
        status_group.add_argument(
            "--kill",
            action = "count",
            dest = "kill",
            help = "Kills the processes of the world. I recommend to use "\
            "the force-stop method if you can. force-stop saves the world, "\
            "before it kills the process."
            )
        status_group.add_argument(
            "--restart",
            action = "count",
            dest = "restart",
            help = "Restarts the world. If the world is offline, "\
            "the world will be started."
            )
        status_group.add_argument(
            "--force-restart",
            action = "count",
            dest = "force_restart",
            help = "Like --restart, but kills the processes to "\
            "stop the world if the smooth stop fails."
            )

        # Setup
        parser.add_argument(
            "--uninstall",
            action = "count",
            dest = "uninstall",
            help = "Uninstalls the world."
            )
        return None

    def run(self, args):
        """
        """
        # We process the worlds in alphabetical order, so we need to sort them.
        worlds = self.app().worlds().get_selected()
        worlds.sort(key = lambda w: w.name())

        for world in worlds:
            world = MyWorld(self.app, world)

            # configuration
            if args.worlds_address:
                world.print_address()
            elif args.configuration:
                world.print_conf()

            elif args.directory:
                world.print_directory()

            # log
            elif args.log is not None\
               or args.log_start is not None \
               or args.log_limit is not None:

                if args.log_start is None:
                    args.log_start = self._default_log_start
                if args.log_limit is None:
                    args.log_limit = self._default_log_limit
                world.print_latest_log(args.log_start, args.log_limit)

            # pid / status / ...
            elif args.pid:
                world.print_pids()

            elif args.status:
                world.print_status()

            # send / screen / ...
            elif args.send:
                world.send_command(args.send)
            elif args.verbose_send:
                world.verbose_send_command(
                    args.verbose_send, self._default_send_command_timeout
                    )
            elif args.console:
                world.open_console(self._default_open_console_delay)

            # start / stop / ...
            elif args.start:
                world.start()
            elif args.stop:
                world.stop()
            elif args.force_stop:
                world.stop(force_stop=True)
            elif args.kill:
                world.kill_processes()
            elif args.restart:
                world.restart()
            elif args.force_restart:
                world.restart(force_restart=True)
            # Setup
            elif args.uninstall:
                world.uninstall()
        return None



#!/usr/bin/env python3

# The MIT License (MIT)
#
# Copyright (c) 2014-2015 Benedikt Schmitt <benedikt@benediktschmitt.de>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
About
-----

Extends the EMSM by a backup manager.

Download
--------

You can find the latest version of this plugin in the EMSM
`GitHub repository <https://github.com/benediktschmitt/emsm>`_.

Configuration
-------------

.. code-block:: ini

    [backups]
    archive_format = bztar
    restore_message = This world is about to be ressetted to an earlier state.
    restore_delay = 5
    max_storage_size = 30
    backup_logs = yes

**archive_format**

    Is the name of the archive format used to create the backups. This string has
    to be listed in *shutil.get_archive_formats()*. Usually, there should be at
    least *zip* or *tar* available.

**restore_message**

    Is send to the world's chat before restoring the world.

**restore_delay**

    Seconds between sending the *restore_message* to the chat and starting
    the restore.

**max_storage_size**

    Maximum number of backups in the storage folder, before older backups
    will be removed.

**backup_logs**

    If ``yes``, the log files are included into the backup, otherwise not.

Arguments
---------

.. note::

    All arguments will only affect the worlds selected with *--world* or
    *--all-world*

.. option:: --list

    Lists all available backups.

.. option:: --create

    Creates a new backup.

.. option:: --restore PATH

    Restores the world with the backup from the given BACKUP_PATH.

.. option:: --restore-latest

    Restores, if available, the latest backup of the world.

.. option:: --restore-menu

    Opens a menu, where the user can select which backup he wants to restore.

Cron
----

You should create a cronjob to create daily backups:

.. code-block:: none

    # m h dom mon dow user command
    # Creates a backup of all worlds everyday at 2:00h
    0 2 *   *   *   root minecraft -W backups --create


Backup archive structure
------------------------

A typical backup archive has this structure:

.. code-block:: none

    o
    |- world_conf.json        # The EMSM configuration of the world
    |- world                  # the minecraft world
        |- server.log
        |- server.properties
        |- ...

Changelog
---------

EMSM v3
^^^^^^^

    * changed package structure and dropped support for EMSM v2 backups.
"""


# Modules
# ------------------------------------------------

# std
import hashlib
import os
import time
import shutil
import datetime
import tempfile
import logging
import json

# third party
import termcolor

# local
import emsm
from emsm.core.base_plugin import BasePlugin


# Backward compatibility
# ------------------------------------------------

try:
    FileExistsError
except NameError:
    FileExistsError = OSError


# Data
# ------------------------------------------------

PLUGIN = "Backups"

AVLB_ARCHIVE_FORMATS = [name for name, desc in shutil.get_archive_formats()]

log = logging.getLogger(__file__)


# Functions
# ------------------------------------------------

def file_hash(path):
    """
    Returns the sha512 hash sum of the file at *path*.
    """
    with open(path, "rb") as file:
        data = file.read()
        sum_ = hashlib.sha512(data)
    return sum_.hexdigest()


# Classes
# ------------------------------------------------

class BackupManager(object):
    """
    Manages the backups of one world.
    """

    def __init__(self, app, world, max_storage_size, backup_dir, backup_logs):
        """
        """
        self._app = app
        self._world = world
        self._backup_dir = backup_dir
        self._max_storage_size = max_storage_size
        self._backup_logs = backup_logs

        os.makedirs(self._backup_dir, exist_ok=True)
        return None

    def app(self):
        """
        Returns that EMSM application that implicitly owns this object.
        """
        return self._app

    def world(self):
        """
        Returns the world whichs backups are managed.
        """
        return self._world

    def backup_dir(self):
        """
        Returns the backup directory.
        """
        return self._backup_dir

    def max_storage_size(self):
        """
        Returns the maximum number of backups that can be stored to the same
        time. If this value is *0*, then there is no limit.
        """
        return self._max_storage_size

    def backup_logs(self):
        """
        Returns the option to include the logs in the backup.
        """
        return self._backup_logs

    # We use the *filenames* to store the *timestamp* of a backup.

    def _filename_format(self):
        """
        Returns a string that can be formatted with *datetime.strftime()*.

        Example:
            >>> now = datetime.datetime.now()
            >>> fmt = bm.filename_format()
            >>> fmt
            %Y_%m_%d-%H_%M_%S-foobar
            >>> now.strftime(fmt)
            '2014_09_02-20_37_08-foobar'
        """
        filename_format = "%Y_%m_%d-%H_%M_%S-{}".format(self._world.name())
        return filename_format

    def _date_from_filename(self, path):
        """
        Extracts the timestamp from the filename and returns the corresponding
        ``datetime.datetime`` instance.

        See also:
            * filename_format()
        """
        # Extract the filename and remove any extensions.
        path = os.path.basename(path)
        filename = path[:path.find(".")]

        # Extract the datetime.
        try:
            return datetime.datetime.strptime(filename, self._filename_format())
        except ValueError as err:
            return None

    def _create_filename(self, date):
        """
        Creates the filename (without extensions) for the backup that has
        been created at the datetime *date*.
        """
        filename = date.strftime(self._filename_format())
        return filename

    def backup_list(self):
        """
        Returns a dictionary that maps the creation date of the backup to
        the backup path.
        """
        backups = dict()
        for filename in os.listdir(self._backup_dir):
            path = os.path.join(self._backup_dir, filename)

            if not os.path.isfile(path):
                continue
            if path.endswith(".tmp"):
                continue

            date = self._date_from_filename(filename)
            if date is None:
                continue

            backups[date] = path
        return backups

    def latest_backup(self):
        """
        Returns a two tuple, that contains the date and the path of the latest
        available backup. If no backup is available, ``(None, None)`` is
        returned.

        See also:
            * backup_list()
        """
        backups = self.backup_list()
        if backups:
            date = max(backups.keys())
            path = backups[date]
            return (date, path)
        else:
            return (None, None)

    def clean_backup_dir(self):
        """
        Removes old backups that are no longer needed.

        See also:
            * max_storage_size()
        """
        # Remove some old backups if we store currently too many backups.
        if self._max_storage_size > 0:
            backups = list(self.backup_list().items())
            backups.sort(reverse=True)

            while len(backups) > self._max_storage_size:
                date, path = backups.pop()
                os.remove(path)

        # Remove .tmp files.
        # These are backups which could not be craeated successfully.
        for filename in os.listdir(self._backup_dir):
            path = os.path.join(self._backup_dir, filename)

            if path.endswith(".tmp"):
                try:
                    os.remove(path)
                except OSError:
                    pass
        return None

    def _save_world(self, backup_dir):
        """
        Copies the world directory (world data) into the backup directory:

            EMSM_ROOT/worlds/foo -> backup_dir/world
        """
        try:
            # We need to disable the auto-save for the backup. I'm paranoid,
            # so I'disable auto-save in this try-catch construct.
            if self._world.is_online():
                self._world.send_command("save-off")
                try:
                    # We use verbose send, to wait until the world has been
                    # saved.
                    self._world.send_command_get_output("save-all", timeout=30)
                except emsm.core.worlds.WorldCommandTimeout as err:
                    pass

            # Copy the world data to *backup_dir*.
            if self._backup_logs:
                shutil.copytree(
                    self._world.directory(),
                    os.path.join(backup_dir, "world")
                )
            else:
                shutil.copytree(
                    self._world.directory(),
                    os.path.join(backup_dir, "world"),
                    ignore=shutil.ignore_patterns("logs")
                )
        finally:
            if self._world.is_online():
                self._world.send_command("save-on")
                self._world.send_command("save-all")
        return None

    def _restore_world(self, backup_dir):
        """
        Copies the world directory from the backup in *backup_dir* into the
        EMSM world folder.

        Exceptions:
            * WorldIsOnlineError

        See also:
            * _save_world()
        """
        # Break if the world is currently online.
        if self._world.is_online():
            raise emsm.core.worlds.WorldIsOnlineError()

        # Delete the world directory (``EMSM/world/...``)
        for i in range(100):
            # XXX: Fixes an error with server.log.lck
            # and shutil.rmtree(...) when the server was online.
            try:
                shutil.rmtree(self._world.directory())
            except OSError:
                time.sleep(0.05)
            else:
                break

        # Copy the world backup to the EMSM world directory.
        shutil.copytree(
            src = os.path.join(backup_dir, "world"),
            dst = self._world.directory()
            )
        return None

    def _save_world_conf(self, backup_dir):
        """
        Saves the configuration of the world in *backup_dir/conf/world.json*.
        """
        world_conf_backup_path = os.path.join(backup_dir, "world_conf.json")
        with open(world_conf_backup_path, "w") as file:
            conf = dict(self._world.conf())
            json.dump([self._world.name(), conf], file)
        return None

    def _restore_world_conf(self, backup_dir):
        """
        If the backup at *backup_dir* includes the world configuration, it
        will be restored. If not, nothing happens.
        """
        world_conf_backup_path = os.path.join(backup_dir, "world_conf.json")

        # Load the configuration backup.
        with open(world_conf_backup_path) as file:
            _, backup_conf = json.load(file)

        # Restore the configuration.
        #
        # Todo:
        #   * restore the *port* option?
        conf = self._world.conf()
        conf.clear()
        conf.update(backup_conf)
        return None

    def create(self, archive_format):
        """
        Creates a backup of the world and returns the name of the created
        backup archive.

        Parameters:
            * archive_format
                A string in shutil.get_archive_formats() that defines the
                compression type.

        Exceptions:
            * ...
        """
        with tempfile.TemporaryDirectory() as tmp_data_dir:

            # Copy all stuff that should be included into the backup in the
            # temporary directory.
            self._save_world(tmp_data_dir)
            self._save_world_conf(tmp_data_dir)

            # Put all in an archive.
            backup_filename = self._create_filename(datetime.datetime.now())
            with tempfile.TemporaryDirectory() as tmp_archive_dir:

                # *make_archive* returns the **complete** path to the crated
                # archive.
                backup_path = shutil.make_archive(
                    base_name = os.path.join(tmp_archive_dir, backup_filename),
                    format = archive_format,
                    root_dir = tmp_data_dir,
                    base_dir = "./"
                    )

                # Move the backup to our folder in *plugins_data_dir*:
                #   EMSM_ROOT/plugins_data/backups/foo/
                #
                # We move the backup to a temporary filename, so that
                # when something goes wrong, no corrupted backup will be
                # stored.
                # When the *move* was successful, we rename the file.
                dst = os.path.join(
                    self._backup_dir, os.path.basename(backup_path)
                    )
                shutil.move(src=backup_path, dst=dst + ".tmp")
                os.rename(dst + ".tmp", dst)

        self.clean_backup_dir()
        return None

    def restore(self, backup_file, message=str(), delay=0):
        """
        Restores the backup of the world from the given *backup_file*. If
        the backup archive contains the server executable it will be restored
        too if necessairy.

        Exceptions:
            * WorldStartFailed
            * WorldStopFailed
            * ... shutil.unpack_archive() exceptions ...
        """
        # Extract the backup in a temporary directory and copy then all things
        # into the EMSM directories.
        with tempfile.TemporaryDirectory() as temp_dir:
            shutil.unpack_archive(
                filename = backup_file,
                extract_dir = temp_dir
                )

            # Stop the world.
            was_online = self._world.is_online()
            if was_online:
                self._world.send_command("say {}".format(message))
                time.sleep(delay)
                self._world.kill_processes()

            # Restore the world.
            self._restore_world(temp_dir)
            self._restore_world_conf(temp_dir)

        # Restart the world if it was online before restoring.
        if was_online:
            self._world.start()
        return None


class UiBackupManager(BackupManager):

    def list(self):
        """
        Prints a list with all existing backups.
        """
        backups = list(self.backup_list().items())
        backups.sort(reverse=True)

        print(termcolor.colored("{}:".format(self.world().name()), "cyan"))
        if not backups:
            print("\t", "- no backups found -")
        else:
            for date, path in backups:
                print("\t", "*", date.ctime())
        return None

    def create(self, archive_format):
        """
        Creates a new backup.
        """
        print(termcolor.colored("{}:".format(self.world().name()), "cyan"))

        try:
            super().create(archive_format)
        except Exception as err:
            # The EMSM will print the exception, so we do nothing here.
            # We only catch Exception to show, that a lot of things
            # can go wrong when creating a backup.
            raise
        else:
            print("\t", "done.")
        return None

    def _restore(self, *, backup_path, message, delay, verify_restore=True):
        """
        The main purpose of this method is simply to wrap the restore
        progress and print it in a user friendly way to the console.

        The *message* and *delay* parameter are passed to the super class
        *restore()* method.

        If *verify_restore* is True, the user is asked if he really wants to
        restore (and so overwrite) the world.
        """
        if verify_restore:
            prompt = "\t Do you really want to restore and " +\
                     termcolor.colored("overwrite", "red") +\
                     " the world '{}'?"
            prompt = prompt.format(self.world().name())
            if not emsm.core.lib.userinput.ask(prompt):
                return None

        # Restore the world.
        try:
            super().restore(backup_path, message, delay)
        except emsm.core.worlds.WorldStopFailed:
            print("\t", termcolor.colored("error:", "red"),
                  "the world could not be stopped.")
        except emsm.core.worlds.WorldStartFailed:
            print("\t", termcolor.colored("error:", "red"),
                  "the world could not be restarted.")
        except Exception as err:
            # This is an unexpected exception that may be an IOError or
            # OSError. Let the EMSM log it and print a warning.
            # We can't handle them properly, since we can not react on
            # them.
            raise
        else:
            print("\t", "done.")
        return None

    def restore(self, backup_path, message, delay, backup_date=None):
        """
        This method corresponds to the command line argument:

            --restore PATH
        """
        print(termcolor.colored("{}:".format(self.world().name()), "cyan"))

        # Print some information about the backup archive.
        print("\t", "backup path: {}".format(backup_path))

        # Restore the backup.
        self._restore(
            backup_path=backup_path, message=message, delay=delay,
            verify_restore=True
            )
        return None

    def restore_latest(self, message, delay):
        """
        This method corresponds to the command line argument:

            --restore-latest
        """
        print(termcolor.colored("{}:".format(self.world().name()), "cyan"))

        # Break if no backup is available.
        latest_backup = self.latest_backup()
        if latest_backup == (None, None):
            print("\t", termcolor.colored("error:", "red"), "no backup available.")
        else:
            date, path = latest_backup
            print("\t", "backup date:", date.ctime())

            # Restore the backup.
            self._restore(
                backup_path=path, message=message, delay=delay,
                verify_restore=True
                )
        return None

    def restore_menu(self, message, delay):
        """
        This method corresponds to the command line argument:

            --restore-menu
        """
        print(termcolor.colored("{}:".format(self.world().name()), "cyan"))

        # Get all existing backups.
        backups = list(self.backup_list().items())
        backups.sort(reverse=True)

        # Break if no backup is available.
        if not backups:
            print("\t", termcolor.colored("error:", "red"),
                  "no backup available.")
        else:
            # Print the backup list.
            for i, backup in enumerate(backups):
                date, path = backup
                print("\t", "{})".format(i), date.ctime())

            # Let the user select a backup.
            i = emsm.core.lib.userinput.get_value(
                prompt = "\t Which backup do you want to restore?",
                conv_func = lambda s: int(s.strip()),
                check_func = lambda s: 0 <= s < len(backups)
                )
            backup = backups[i]

            # Restore the backup.
            self._restore(
                backup_path=backup[1], message=message, delay=delay,
                verify_restore=True
                )
        return None


class Backups(BasePlugin):

    VERSION = "4.0.0-beta"

    DESCRIPTION = __doc__

    def __init__(self, app, name):
        """
        """
        BasePlugin.__init__(self, app, name)

        self._setup_conf()
        self._setup_argparser()
        return None

    def _setup_conf(self):
        """
        Sets the configuration up.
        """
        conf = self.conf()

        # Read
        # ^^^^

        # archive_format
        self._archive_format = conf.get("archive_format")
        if not self._archive_format in AVLB_ARCHIVE_FORMATS:
            self._archive_format = AVLB_ARCHIVE_FORMATS[0]

        # restore_message
        self._restore_message = conf.get(
            "restore_message",
            "This world is about to be resetted to an earlier state."
            )

        # restore_delay
        self._restore_delay = conf.getint("restore_delay", 5)
        if self._restore_delay < 0:
            self._restore_delay = 0

        # max_storage_size
        self._max_storage_size = conf.getint("max_storage_size", 30)
        if self._max_storage_size < 0:
            self._max_storage_size = 0

        # backup_logs
        self._backup_logs = conf.getboolean("backup_logs", True)

        # Write
        # ^^^^^

        conf.clear()
        conf["archive_format"] = str(self._archive_format)
        conf["restore_message"] = str(self._restore_message)
        conf["restore_delay"] = str(self._restore_delay)
        conf["max_storage_size"] = str(self._max_storage_size)
        conf["backup_logs"] = "yes" if self._backup_logs else "no"
        return None

    def _setup_argparser(self):
        """
        Sets the argument parser up.
        """
        parser = self.argparser()

        parser.description = "A manager for the backups of the worlds"

        # We want to allow only one action per run, so we put everything
        # in a mutually exclusive group.
        me_group = parser.add_mutually_exclusive_group()
        me_group.add_argument(
            "--list",
            action = "count",
            dest = "backups_list",
            help = "Lists all available backups."
            )
        me_group.add_argument(
            "--create",
            action = "count",
            dest = "backups_create",
            help = "Creates a new backup."
            )
        me_group.add_argument(
            "--restore",
            action = "store",
            dest = "backups_restore",
            metavar = "PATH",
            help = "Restores the backup at the given path."
            )
        me_group.add_argument(
            "--restore-latest",
            action = "count",
            dest = "backups_restore_latest",
            help = "Restores the latest backup."
            )
        me_group.add_argument(
            "--restore-menu",
            action = "count",
            dest = "backups_restore_menu",
            help = "Opens a dialog allowing the user to select the backup "\
                   "that should be restored."
            )
        return None

    def run(self, args):
        """
        """
        # We process the worlds in alphabetical order, so we need to sort
        # them first.
        worlds = self.app().worlds().get_selected()
        worlds.sort(key = lambda w: w.name())

        for world in worlds:
            bm = UiBackupManager(
                app = self.app(),
                world = world,
                max_storage_size = self._max_storage_size,
                backup_dir = os.path.join(self.data_dir(), world.name()),
                backup_logs = self._backup_logs
                )

            if args.backups_list:
                bm.list()
            elif args.backups_create:
                bm.create(self._archive_format)
            elif args.backups_restore:
                bm.restore(args.backups_restore, self._restore_message,
                           self._restore_delay
                           )
            elif args.backups_restore_latest:
                bm.restore_latest(self._restore_message, self._restore_delay)
            elif args.backups_restore_menu:
                bm.restore_menu(self._restore_message, self._restore_delay)
        return None
